// TODO: Should there be callbacks for monitors? UBUS says no, but APB and VIP say yes..

class basic_monitor #(type VIF = int, type T = basic_tr) extends uvm_monitor;
  typedef basic_monitor #(VIF, T) this_t;

  VIF  vif;

  uvm_analysis_port #(T) item_collected_port;
  
  // It is optional to have protocol checking. We will
  // avoid it here as dpcm_if does the needed checking depends on
  // the precision of the signal sent over the bus which is only known on a
  // scenario level (i.e. a level above this monitor). Fields exist for forward
  // compatibility.
  bit checks_enable = 1; //TODO: Why is the naming different from the ifs?

  // Latency through the LPCM sample pipeline of the DUT. This is only relevant
  // to set for downstream monitors so that a scoreboard will not detect a
  // mismatch on the very first received sample out of reset. This should not
  // be set for the master monitor.
  int unsigned pipeline_latency;
  
  `uvm_component_param_utils_begin(this_t)
    `uvm_field_int(checks_enable, UVM_DEFAULT)
    `uvm_field_int(pipeline_latency, UVM_DEFAULT)
  `uvm_component_utils_end
    
  function new(string name, uvm_component parent);
    super.new(name, parent);
    item_collected_port = new ("item_collected_port", this);
    pipeline_latency = 0;
  endfunction

  virtual function void build_phase(uvm_phase phase);
    super.build_phase(phase);
    if (!uvm_config_db#(VIF)::get(this, "", "vif", vif)) begin
      `uvm_error("NOVIF", {"virtual inferface must be set for: ", get_full_name(), ".vif"})
    end
  endfunction

  virtual task run_phase(uvm_phase phase);
    int latency;
    super.run_phase(phase);
    latency = -1-pipeline_latency;
    wait(vif.resetb === 0);
    @(posedge vif.resetb);
    `uvm_info(get_type_name(), $sformatf("Latency after reset: %0d", latency), UVM_HIGH)
    @(posedge vif.clk);
    forever begin
      T tr;
      while (vif.en === 0) begin
        @(posedge vif.clk);
        latency += 1;
      end;
      phase.raise_objection(this,"Processing LPCM sample");
      tr = T::type_id::create("tr", this);
      assert(begin_tr(tr, "basic_monitor"));  // for transaction recording
      tr.sample = vif.data;
      tr.latency = latency;
      @(posedge vif.clk);
      end_tr(tr);

     `uvm_info(get_type_name(), $sformatf("Writing latency %0d", latency), UVM_HIGH)

      if (checks_enable) perform_checks();
      collect_coverage(); 
      
      item_collected_port.write(tr);
      phase.drop_objection(this);      

      latency = 0;
    end
  endtask

  function void perform_checks();
    // This is a placeholder function. Its purpose is to show where we would
    // have put checks of the sampled data, if we had anything to check on a
    // transaction level in addition to the checks we already performed on the
    // signal/interface level. Unfortunately we don't know enough information
    // about the data on this level to perform the final checks, so we have to
    // defer this to the scenario level.
  endfunction

  function void collect_coverage();
    // This is a placeholder function. Its purpose is to show where we would
    // have put coverage collection if we had known enough about our data to
    // perform a reliable collection at this level. Unfortunately we don't know
    // enough about the received data, so we have to defer the coverage
    // collection to a higher (scenario) level.
  endfunction
endclass
