//
// A Functional Coverage collection library for LPCM
//
// We want to cover the following conditions which we have found through a
// quick domain analysis. Input and output domains are the same which makes
// everything easier:
//
// Equivalence Partitioning:
//  - 16 bit, 24 bit and 32 bit precision
//  - Positive, negative numbers, and zero
//
// Special values in fixed point arithmetic operations
//  - Addition: 0
//  - Multiplication: 0, 1.0 (doesn't exist), -1.0 == Minimum
//  - Division: Zero, Near Zero (-eps, eps)
//  - Saturation: Maximum, Minimum
//
// Knowledge about error history in serialization of LPCM
//  - Occasionally wrong LSB
//
// Boundary Value Analysis of all of the above
//  - Maximum, Minimum, Maximum-1, Minimum+1, -1, 0, 1
//
// In the current incarnation of this verification environment the system clock
// rate is a secondary concern. In a real world example we would most likely
// include the system clock and latency to give us the sample rate of the LPCM
// transactions such that we could verify it and also assert that we have seen
// all supported sample rates which the system supports.
//
// Finally we want to make sure that unused LSBs infact are zero.

parameter LPCM_MAXN = 1<<(`LPCM_SIZE-1);

`include "bvm_cover.sv"

class lpcm_cover extends bvm_cover #(lpcm_tr);
  bit checks_enable = 1;

  `uvm_component_utils_begin(lpcm_cover)
    `uvm_field_int(checks_enable, UVM_DEFAULT)
  `uvm_component_utils_end

  lpcm_tr tra;
  covergroup cov_tr;
    option.per_instance = 1;
    // We are not interested in latency
    sample_cp : coverpoint tra.sample {
      bins special[3] = {-1, 0, 1};
      bins positive   = {[2:LPCM_MAXN-3]};
      bins maximum[2] = {LPCM_MAXN-2, LPCM_MAXN-1};
      bins negative   = {[-LPCM_MAXN+2:-2]};
      bins minimum[2] = {-LPCM_MAXN, -(LPCM_MAXN-1)};
      option.at_least = 3;
    }
    LSBs : coverpoint {tra.sample[31], tra.sample[1:0]} {
      bins pos_lsbs[4] = {[0:3]};
      bins neg_lsbs[4] = {[4:7]};
      option.at_least = 1000;
    }
  endgroup

  function new(string name, uvm_component parent);
    super.new(name, parent);
    cov_tr = new();
    cov_tr.set_inst_name({get_full_name(), ".cov_tr"});
  endfunction
  
  `bvm_cover_utils(lpcm_tr, tra, cov_tr)

endclass

